//==========================================================
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF 
// ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO 
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A 
// PARTICULAR PURPOSE.
//
// (C)Copyright 1995, Microsoft Corporation. All rights reserved.
//
// This is a minimal application designed as a demonstration vehicle
// for a simple VxD. This application simply loads the VxD, issues
// a DeviceIOControl call to it, and waits for an APC from the VxD.
//
// NOTE: Notice the main message loop. It is important that while
// waiting for APCs, and app does not stop processing messages. If
// an app does stop processing messages, it can potentially hang
// another application that does a BroadcastSystemMessage.
//==========================================================
 

#include <windows.h>
#include "makethrd.h"

// DeviceIOControl Codes
// Macro definition for defining NT compatible IOCTL and FSCTL function
// control codes. Note that function codes 0-2047 are reserved for
// Microsoft Corporation, and 2048-4095 are reserved for customers.
//
// Note that DeviceType values used by Microsoft
// Corporation are in the range 0-32767, and 32768-65535 are reserved for use
// by customers.
//
// Method 
#define METHOD_NEITHER          3
#define FILE_ANY_ACCESS         0

#define CTL_CODE( DeviceType, Function, Method, Access ) (              \
    ((DeviceType) << 16) | ((Access) << 14) | ((Function) << 2) | (Method))

#define DEVICE_WIN95VXD         32768

// Constant defines
#define SLEEP_DELAY 100	                // wait for 1/10 second between peeking for msg

// code to get version
#define MAKETHRD_GetVersion CTL_CODE( DEVICE_WIN95VXD, 2048, METHOD_NEITHER, FILE_ANY_ACCESS )

// code to create a thread
#define MAKETHRD_MakeThread CTL_CODE( DEVICE_WIN95VXD, 2049, METHOD_NEITHER, FILE_ANY_ACCESS )

// Prototypes for B-list APIs used
VOID * APIENTRY OpenVxDHandle(HANDLE hnd);

// Structure & types for communication with the VxD

typedef struct _threadstartstruct {
  HANDLE	hRing0Event;		// VxDHandle to Win32 event
  HANDLE	hRing3Event;		// parameter for ring 3 function
  HANDLE	hVxD;				// handle to the VxD itself
  HANDLE	hRing0Thread;		// handle to the ring 0 thread
  LPVOID	pfnRing3APC;		// function to call at ring 3
  LPVOID	thcbUserAPC;		// thread to queue user APC to (used only by ring 0)
} THREAD_START_STRUCT;


// Strings
char szAppName[] = "MAKETHRD";
char szVxDName[] = "\\\\.\\MAKETHRD.VXD";
char szTitle[]   = "Windows 95 VxD Example Application";

char szErrMsg0[] = "VxD load error";
char szErrMsg1[] = "VxD does not support DeviceIOControl";
char szErrMsg2[] = "Unable to open VxD";
char szErrMsg3[] = "Unable to create Event";
char szErrMsg4[] = "An error occurred in the VxD while servicing the request";

char szAPCMsg0[] = "Message from the ring 0 thread...";
char szAPCMsg1[] = "I'm alive!";

// Prototypes
BOOL    InitApplication(HINSTANCE hInstance, int nCmdShow);
LRESULT CALLBACK WndProc(HWND hwnd, UINT uiMessage, WPARAM wParam, LPARAM lParam);
VOID	APIENTRY APCFunction( THREAD_START_STRUCT *ptst );
BOOL	LoadVxD(VOID);

//  WinMain
int	APIENTRY WinMain(	HINSTANCE hInstance,
						HINSTANCE hPrevInstance,
						LPSTR lpCmdLine,
						int nCmdShow)
{
    MSG msg;
	BOOL fQuit = FALSE;

    // Initialize classes, etc.
    if (!InitApplication(hInstance, nCmdShow))
        return FALSE;

    // Main message loop. Since there isn't a way to allow APCs to be delivered
	// as well as messages, we actually stay in a peekmessage loop, and between
	// peekmessages, we sleep our thread for a defined period of time. This keeps
	// us from running all the time doing peekmesage and allows message processing
	// as well as delivering of apcs.
	do {
		while (!PeekMessage(&msg, NULL, 0, 0xffffffff, PM_REMOVE)) {
			SleepEx( SLEEP_DELAY, TRUE );
		}

		if (msg.message != WM_QUIT) {
			// translate and dispatch the message
        	TranslateMessage(&msg);
        	DispatchMessage(&msg);
		}

	} while (msg.message != WM_QUIT);

    return msg.wParam;
}


//  InitApplication
BOOL InitApplication( HINSTANCE hInstance, int nCmdShow )
{
    WNDCLASSEX wc;
    HWND hwnd;

    wc.cbSize = sizeof(WNDCLASSEX);
    wc.style = 0;
    wc.lpfnWndProc = (WNDPROC)WndProc;
    wc.cbClsExtra = 0;
    wc.cbWndExtra = 0;
    wc.hInstance = hInstance;
    wc.hIcon = LoadIcon (hInstance, MAKEINTRESOURCE(ID_ICON));
    wc.hCursor = LoadCursor(NULL, IDC_ARROW);
    wc.hbrBackground = (HBRUSH)(COLOR_WINDOW + 1);
    wc.lpszMenuName = MAKEINTRESOURCE(ID_MENU);
    wc.lpszClassName = szAppName;
    wc.hIconSm = NULL;

    if (!RegisterClassEx(&wc))
        return FALSE;

    hwnd = CreateWindow(szAppName, szTitle,
        WS_OVERLAPPEDWINDOW,
        CW_USEDEFAULT, 0, CW_USEDEFAULT, 0,
        NULL, NULL, hInstance, NULL);

    if (!hwnd)
        return FALSE;

    ShowWindow(hwnd, nCmdShow);
    UpdateWindow(hwnd);

    return TRUE;
}

//  WndProc
LRESULT	CALLBACK WndProc( HWND hwnd, UINT uiMessage, WPARAM wParam, LPARAM lParam)
{
    switch (uiMessage)
    {
    	case WM_COMMAND:
        	switch (wParam)
        	{
				case IDM_LOAD:
				    LoadVxD();
				break;

        		case IDM_EXIT:
            		SendMessage(hwnd, WM_CLOSE, 0, 0L);
            	break;

        	}
        	break;

    	case WM_CLOSE:
        	DestroyWindow(hwnd);
        	PostQuitMessage(0);
		break;

    	default:
        	return DefWindowProc(hwnd, uiMessage, wParam, lParam);
    }

    return 0;
}


// This function actually loads and communicates with the VxD. If it
// isn't successful in loading the VxD, it returns FALSE.
BOOL LoadVxD(VOID)
{
	THREAD_START_STRUCT tst;
	DWORD cbBytesReturned;
	HANDLE hNewThread;

	// Load the service VxD.
    tst.hVxD = CreateFile(szVxDName, 0,0,0,	CREATE_NEW, FILE_FLAG_DELETE_ON_CLOSE, 0);

	// if we had a load error, bail
    if (tst.hVxD != INVALID_HANDLE_VALUE) {

	    // Now, we'll set up the ring 3 event, and get a VxD handle to it
		if (tst.hRing3Event = CreateEvent( NULL, FALSE, FALSE, NULL )) {
		    // open a handle accessable from ring 0
		    tst.hRing0Event = OpenVxDHandle( tst.hRing3Event );
			// point to APC function
			tst.pfnRing3APC = &APCFunction;

    		if (DeviceIoControl(tst.hVxD, (DWORD)MAKETHRD_MakeThread, (LPVOID)(&tst), sizeof(tst),
                        		(LPVOID)(&hNewThread), sizeof(DWORD), &cbBytesReturned, NULL)) {
				// one last sanity check
				if (cbBytesReturned == sizeof(DWORD)) {
					return TRUE;
				}
			}
			
			// if we get here, the VxD had an error trying to service the request
			MessageBox(NULL, szErrMsg4,	szErrMsg0, MB_ICONEXCLAMATION | MB_OK);

		} else {
			// we had an error creating the event
			MessageBox(NULL, szErrMsg3, szErrMsg0, MB_ICONEXCLAMATION | MB_OK);
		}

		// Once we get here, we can close the VxD handle
		CloseHandle(tst.hVxD);

	} else {
        if (GetLastError() == ERROR_NOT_SUPPORTED) {
            MessageBox(NULL, szErrMsg1, szErrMsg0, MB_ICONEXCLAMATION | MB_OK);
        } else {
            MessageBox(NULL, szErrMsg2, szErrMsg0, MB_ICONEXCLAMATION | MB_OK);
        }
    }
    
	// we have a return in the middle of this function which we hit on success
	return FALSE;
}

// This is the function which executes as an APC. It is queued by the
// VxD and passed the ring 3 event handle which was originally allocated
// here at ring 3. When we get here, we will put up a message box. When
// the message box is dismissed, we signal the event, close the handle,
// and return.
//
// Although the structure we are passed looks like the one we passed to the
// VxD, it has been copied to a new block of memory by the VxD. This way
// the VxD can use it in the new thread which is not in our process context.
VOID APIENTRY APCFunction( THREAD_START_STRUCT *ptst )
{
	// display success to the user
	MessageBox(NULL, szAPCMsg1, szAPCMsg0, MB_ICONEXCLAMATION | MB_OK );

	// signal to the ring 0 thread that we did it
	SetEvent( ptst->hRing3Event );
	
	// close the event handle, it won't go away until the ring 0 thread
	// also closes its VxD handle to the event
	CloseHandle( ptst->hRing3Event );

	// wait until the thread terminates so that we don't close the VxD
	// while it has a thread in it.
	WaitForSingleObject( ptst->hRing0Thread, INFINITE );
	CloseHandle( ptst->hRing0Thread );

	// close the VxD handle and we're done
	CloseHandle( ptst->hVxD );
}
